@@ -89,11 +89,15 @@ syscall_openat2 (int dirfd, const char *path, uint64_t flags, uint64_t mode, uin
 }
 
 int
-crun_path_exists (const char *path, libcrun_error_t *err arg_unused)
+crun_path_exists (const char *path, libcrun_error_t *err)
 {
   int ret = access (path, F_OK);
   if (ret < 0)
-    return errno == ENOENT ? 0 : ret;
+    {
+      if (errno == ENOENT)
+        return 0;
+      return crun_make_error (err, errno, "access `%s`", path);
+    }
   return 1;
 }
 
@@ -116,7 +120,7 @@ xasprintf (char **str, const char *fmt, ...)
 int
 write_file_at_with_flags (int dirfd, int flags, mode_t mode, const char *name, const void *data, size_t len, libcrun_error_t *err)
 {
-  cleanup_close int fd = openat (dirfd, name, O_WRONLY | flags, mode);
+  cleanup_close int fd = openat (dirfd, name, O_CLOEXEC | O_WRONLY | flags, mode);
   int ret = 0;
   if (UNLIKELY (fd < 0))
     return crun_make_error (err, errno, "opening file `%s` for writing", name);
@@ -134,13 +138,13 @@ write_file_at_with_flags (int dirfd, int flags, mode_t mode, const char *name, c
 int
 write_file_at (int dirfd, const char *name, const void *data, size_t len, libcrun_error_t *err)
 {
-  return write_file_at_with_flags (dirfd, O_CREAT | O_TRUNC, 0700, name, data, len, err);
+  return write_file_at_with_flags (dirfd, O_CLOEXEC | O_CREAT | O_TRUNC, 0700, name, data, len, err);
 }
 
 int
 write_file_with_flags (const char *name, int flags, const void *data, size_t len, libcrun_error_t *err)
 {
-  cleanup_close int fd = open (name, O_WRONLY | flags, 0700);
+  cleanup_close int fd = open (name, O_CLOEXEC | O_WRONLY | flags, 0700);
   int ret;
   if (UNLIKELY (fd < 0))
     return crun_make_error (err, errno, "opening file `%s` for writing", name);
@@ -819,19 +823,19 @@ add_selinux_mount_label (char **retlabel, const char *data, const char *label, c
   return 0;
 }
 
-static int
-set_security_attr (const char *lsm, const char *fname, const char *data, libcrun_error_t *err)
+static const char *
+lsm_attr_path (const char *lsm, const char *fname, libcrun_error_t *err)
 {
-  int ret;
-  struct statfs sfs;
-
   cleanup_close int attr_dirfd = -1;
   cleanup_close int lsm_dirfd = -1;
-  cleanup_close int fd = -1;
+  char *attr_path = NULL;
 
   attr_dirfd = open ("/proc/thread-self/attr", O_DIRECTORY | O_RDONLY | O_CLOEXEC);
   if (UNLIKELY (attr_dirfd < 0))
-    return crun_make_error (err, errno, "open `/proc/thread-self/attr`");
+    {
+      crun_make_error (err, errno, "open `/proc/thread-self/attr`");
+      return NULL;
+    }
 
   // Check for newer scoped interface in /proc/thread-self/attr/<lsm>
   if (lsm != NULL)
@@ -839,34 +843,56 @@ set_security_attr (const char *lsm, const char *fname, const char *data, libcrun
       lsm_dirfd = openat (attr_dirfd, lsm, O_DIRECTORY | O_RDONLY | O_CLOEXEC);
 
       if (UNLIKELY (lsm_dirfd < 0 && errno != ENOENT))
-        return crun_make_error (err, errno, "open `/proc/thread-self/attr/%s`", lsm);
+        {
+          crun_make_error (err, errno, "open `/proc/thread-self/attr/%s`", lsm);
+          return NULL;
+        }
     }
-
   // Use scoped interface if available, fall back to unscoped
-  if (lsm_dirfd >= 0)
-    fd = openat (lsm_dirfd, fname, O_WRONLY | O_CLOEXEC);
-  else
-    fd = openat (attr_dirfd, fname, O_WRONLY | O_CLOEXEC);
+  xasprintf (&attr_path, "/proc/thread-self/attr/%s%s%s", lsm_dirfd >= 0 ? lsm : "", lsm_dirfd >= 0 ? "/" : "", fname);
 
-  if (UNLIKELY (fd < 0))
-    return crun_make_error (err, errno, "open `/proc/thread-self/attr/%s%s%s`",
-                            lsm_dirfd >= 0 ? lsm : "", lsm_dirfd >= 0 ? "/" : "", fname);
+  return attr_path;
+}
 
-  // Check that the file system type is indeed procfs
-  ret = fstatfs (fd, &sfs);
+static int
+check_proc_super_magic (int fd, const char *path, libcrun_error_t *err)
+{
+  struct statfs sfs;
+
+  int ret = fstatfs (fd, &sfs);
   if (UNLIKELY (ret < 0))
-    return crun_make_error (err, errno, "statfs `/proc/thread-self/attr/%s%s%s`",
-                            lsm_dirfd >= 0 ? lsm : "", lsm_dirfd >= 0 ? "/" : "", fname);
+    return crun_make_error (err, errno, "statfs `%s`", path);
 
   if (sfs.f_type != PROC_SUPER_MAGIC)
-    return crun_make_error (err, 0, "the file `/proc/thread-self/attr/%s%s%s` is not on a `procfs` file system",
-                            lsm_dirfd >= 0 ? lsm : "", lsm_dirfd >= 0 ? "/" : "", fname);
+    return crun_make_error (err, 0, "the file `%s` is not on a `procfs` file system", path);
+
+  return 0;
+}
+
+static int
+set_security_attr (const char *lsm, const char *fname, const char *data, libcrun_error_t *err)
+{
+  int ret;
+
+  cleanup_free const char *attr_path = lsm_attr_path (lsm, fname, err);
+  cleanup_close int fd = -1;
+
+  if (UNLIKELY (attr_path == NULL))
+    return -1;
+
+  fd = open (attr_path, O_WRONLY | O_CLOEXEC);
+
+  if (UNLIKELY (fd < 0))
+    return crun_make_error (err, errno, "open `%s`", attr_path);
+
+  ret = check_proc_super_magic (fd, attr_path, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
 
   // Write out data
   ret = TEMP_FAILURE_RETRY (write (fd, data, strlen (data)));
   if (UNLIKELY (ret < 0))
-    return crun_make_error (err, errno, "write file `/proc/thread-self/attr/%s%s%s`",
-                            lsm_dirfd >= 0 ? lsm : "", lsm_dirfd >= 0 ? "/" : "", fname);
+    return crun_make_error (err, errno, "write file `%s`", attr_path);
 
   return 0;
 }
@@ -893,8 +919,33 @@ libcrun_is_apparmor_enabled (libcrun_error_t *err)
   return apparmor_enabled;
 }
 
+static int
+is_current_process_confined (libcrun_error_t *err)
+{
+  cleanup_free const char *attr_path = lsm_attr_path ("apparmor", "current", err);
+  cleanup_close int fd = -1;
+  char buf[256];
+
+  if (UNLIKELY (attr_path == NULL))
+    return -1;
+
+  fd = open (attr_path, O_RDONLY | O_CLOEXEC);
+
+  if (UNLIKELY (fd < 0))
+    return crun_make_error (err, errno, "open `%s`", attr_path);
+
+  if (UNLIKELY (check_proc_super_magic (fd, attr_path, err)))
+    return -1;
+
+  ssize_t bytes_read = read (fd, buf, sizeof (buf) - 1);
+  if (UNLIKELY (bytes_read < 0))
+    return crun_make_error (err, errno, "error reading file `%s`", attr_path);
+
+  return (strncmp (buf, "unconfined", bytes_read) != 0 && buf[0] != '\0');
+}
+
 int
-set_apparmor_profile (const char *profile, bool now, libcrun_error_t *err)
+set_apparmor_profile (const char *profile, bool no_new_privileges, bool now, libcrun_error_t *err)
 {
   int ret;
 
@@ -905,8 +956,13 @@ set_apparmor_profile (const char *profile, bool now, libcrun_error_t *err)
   if (ret)
     {
       cleanup_free char *buf = NULL;
-
-      xasprintf (&buf, "%s %s", now ? "changeprofile" : "exec", profile);
+      ret = is_current_process_confined (err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+      // if confined only way for apparmor to allow change of profile with NNP is with stacking
+      xasprintf (&buf, "%s %s", no_new_privileges && ret ? "stack" : now ? "changeprofile"
+                                                                         : "exec",
+                 profile);
 
       return set_security_attr ("apparmor", now ? "current" : "exec", buf, err);
     }
@@ -981,7 +1037,7 @@ read_all_file_at (int dirfd, const char *path, char **out, size_t *len, libcrun_
 {
   cleanup_close int fd;
 
-  fd = TEMP_FAILURE_RETRY (openat (dirfd, path, O_RDONLY));
+  fd = TEMP_FAILURE_RETRY (openat (dirfd, path, O_RDONLY | O_CLOEXEC));
   if (UNLIKELY (fd < 0))
     return crun_make_error (err, errno, "error opening file `%s`", path);
 
@@ -1012,7 +1068,7 @@ open_unix_domain_client_socket (const char *path, int dgram, libcrun_error_t *er
 
   if (strlen (path) >= sizeof (addr.sun_path))
     {
-      destfd = open (path, O_PATH);
+      destfd = open (path, O_PATH | O_CLOEXEC);
       if (UNLIKELY (destfd < 0))
         return crun_make_error (err, errno, "error opening `%s`", path);
 
@@ -1365,7 +1421,7 @@ set_home_env (uid_t id)
 
   buf = xmalloc (buf_size);
 
-  stream = fopen ("/etc/passwd", "r");
+  stream = fopen ("/etc/passwd", "re");
   if (stream == NULL)
     {
       if (errno == ENOENT)
@@ -1448,7 +1504,7 @@ getsubidrange (uid_t id, int is_uid, uint32_t *from, uint32_t *len)
 
   len_name = strlen (name);
 
-  input = fopen (is_uid ? "/etc/subuid" : "/etc/subgid", "r");
+  input = fopen (is_uid ? "/etc/subuid" : "/etc/subgid", "re");
   if (input == NULL)
     return -1;
 
@@ -1513,6 +1569,18 @@ format_default_id_mapping (char **ret, uid_t container_id, uid_t host_uid, uid_t
   return written;
 }
 
+static int
+unset_cloexec_flag (int fd)
+{
+  int flags = fcntl (fd, F_GETFD);
+  if (flags == -1)
+    return -1;
+
+  flags &= ~FD_CLOEXEC;
+
+  return fcntl (fd, F_SETFD, flags);
+}
+
 static void __attribute__ ((__noreturn__))
 run_process_child (char *path, char **args, const char *cwd, char **envp, int pipe_r,
                    int pipe_w, int out_fd, int err_fd)
@@ -1528,7 +1596,7 @@ run_process_child (char *path, char **args, const char *cwd, char **envp, int pi
 
   if (out_fd < 0 || err_fd < 0)
     {
-      dev_null_fd = open ("/dev/null", O_WRONLY);
+      dev_null_fd = open ("/dev/null", O_WRONLY | O_CLOEXEC);
       if (UNLIKELY (dev_null_fd < 0))
         _exit (EXIT_FAILURE);
     }
@@ -1540,6 +1608,11 @@ run_process_child (char *path, char **args, const char *cwd, char **envp, int pi
   dup2 (out_fd >= 0 ? out_fd : dev_null_fd, 1);
   dup2 (err_fd >= 0 ? err_fd : dev_null_fd, 2);
 
+  if (out_fd >= 0)
+    unset_cloexec_flag (1);
+  if (err_fd >= 0)
+    unset_cloexec_flag (2);
+
   if (dev_null_fd >= 0)
     TEMP_FAILURE_RETRY (close (dev_null_fd));
   if (out_fd >= 0)
@@ -1573,7 +1646,7 @@ run_process_with_stdin_timeout_envp (char *path, char **args, const char *cwd, i
 
   sigemptyset (&mask);
 
-  ret = pipe (stdin_pipe);
+  ret = pipe2 (stdin_pipe, O_CLOEXEC);
   if (UNLIKELY (ret < 0))
     return crun_make_error (err, errno, "pipe");
   pipe_r = stdin_pipe[0];
@@ -1674,7 +1747,6 @@ mark_or_close_fds_ge_than (int n, bool close_now, libcrun_error_t *err)
   cleanup_dir DIR *dir = NULL;
   int ret;
   int fd;
-  struct statfs sfs;
   struct dirent *next;
 
   ret = syscall_close_range (n, UINT_MAX, close_now ? 0 : CLOSE_RANGE_CLOEXEC);
@@ -1687,12 +1759,9 @@ mark_or_close_fds_ge_than (int n, bool close_now, libcrun_error_t *err)
   if (UNLIKELY (cfd < 0))
     return crun_make_error (err, errno, "open `/proc/self/fd`");
 
-  ret = fstatfs (cfd, &sfs);
+  ret = check_proc_super_magic (cfd, "/proc/self/fd", err);
   if (UNLIKELY (ret < 0))
-    return crun_make_error (err, errno, "statfs `/proc/self/fd`");
-
-  if (sfs.f_type != PROC_SUPER_MAGIC)
-    return crun_make_error (err, 0, "the path `/proc/self/fd` is not on file system type `procfs`");
+    return ret;
 
   dir = fdopendir (cfd);
   if (UNLIKELY (dir == NULL))
@@ -2033,11 +2102,11 @@ copy_recursive_fd_to_fd (int srcdirfd, int dfd, const char *srcname, const char
       switch (mode & S_IFMT)
         {
         case S_IFREG:
-          srcfd = openat (dirfd (dsrcfd), de->d_name, O_NONBLOCK | O_RDONLY);
+          srcfd = openat (dirfd (dsrcfd), de->d_name, O_NONBLOCK | O_RDONLY | O_CLOEXEC);
           if (UNLIKELY (srcfd < 0))
             return crun_make_error (err, errno, "open `%s/%s`", srcname, de->d_name);
 
-          destfd = openat (destdirfd, de->d_name, O_RDWR | O_CREAT, 0777);
+          destfd = openat (destdirfd, de->d_name, O_RDWR | O_CREAT | O_CLOEXEC, 0777);
           if (UNLIKELY (destfd < 0))
             return crun_make_error (err, errno, "open `%s/%s`", destname, de->d_name);
 
@@ -2060,11 +2129,11 @@ copy_recursive_fd_to_fd (int srcdirfd, int dfd, const char *srcname, const char
           if (UNLIKELY (ret < 0))
             return crun_make_error (err, errno, "mkdir `%s/%s`", destname, de->d_name);
 
-          srcfd = openat (dirfd (dsrcfd), de->d_name, O_DIRECTORY);
+          srcfd = openat (dirfd (dsrcfd), de->d_name, O_DIRECTORY | O_CLOEXEC);
           if (UNLIKELY (srcfd < 0))
             return crun_make_error (err, errno, "open directory `%s/%s`", srcname, de->d_name);
 
-          destfd = openat (destdirfd, de->d_name, O_DIRECTORY);
+          destfd = openat (destdirfd, de->d_name, O_DIRECTORY | O_CLOEXEC);
           if (UNLIKELY (destfd < 0))
             return crun_make_error (err, errno, "open directory `%s/%s`", srcname, de->d_name);
 
@@ -2118,8 +2187,7 @@ copy_recursive_fd_to_fd (int srcdirfd, int dfd, const char *srcname, const char
           if (errno == ENOTSUP)
             continue;
 
-          if (UNLIKELY (ret < 0))
-            return crun_make_error (err, errno, "chmod `%s/%s`", destname, de->d_name);
+          return crun_make_error (err, errno, "chmod `%s/%s`", destname, de->d_name);
         }
     }
 
